@@ -21,6 +21,8 @@
 
 #include "exslt.h"
 
+#ifdef EXSLT_CRYPTO_ENABLED
+
 #define HASH_DIGEST_LENGTH 32
 #define MD5_DIGEST_LENGTH 16
 #define SHA1_DIGEST_LENGTH 20
@@ -235,7 +237,7 @@ exsltCryptoCryptoApiRc4Encrypt (xmlXPathParserContextPtr ctxt,
 	exsltCryptoCryptoApiReportError (ctxt, __LINE__);
 	goto fail;
     }
-// Now encrypt data.
+/* Now encrypt data. */
     dwDataLen = msglen;
     memcpy (dest, msg, msglen);
     if (!CryptEncrypt (hKey, 0, TRUE, 0, dest, &dwDataLen, msglen)) {
@@ -286,7 +288,7 @@ exsltCryptoCryptoApiRc4Decrypt (xmlXPathParserContextPtr ctxt,
 	exsltCryptoCryptoApiReportError (ctxt, __LINE__);
 	goto fail;
     }
-// Now encrypt data.
+/* Now encrypt data. */
     dwDataLen = msglen;
     memcpy (dest, msg, msglen);
     if (!CryptDecrypt (hKey, 0, TRUE, 0, dest, &dwDataLen)) {
@@ -351,7 +353,8 @@ exsltCryptoGcryptInit (void) {
  */
 static void
 exsltCryptoGcryptHash (xmlXPathParserContextPtr ctxt ATTRIBUTE_UNUSED,
-		       enum gcry_md_algos algorithm, const char *msg,
+/* changed the enum to int */
+		       int algorithm, const char *msg,
 		       unsigned long msglen,
 		       char dest[HASH_DIGEST_LENGTH]) {
     exsltCryptoGcryptInit ();
@@ -731,3 +734,5 @@ exsltCryptoRegister (void) {
 }
 
 #endif /* defined(HAVE_CRYPTO) */
+
+#endif /* EXSLT_CRYPTO_ENABLED */
